public abstract class HiveAI<U extends AIInterface<W>,W> implements ServerListener {

	/**The server handler which is initialized in Main.*/
	private ServerHandler handler;
	/**The ServerCommand which is initialized in awaitLogin.*/
	protected ServerCommand command;
	/***/
	private String username;
	private String password;
	private String loginName;
	protected PlayerColor playerColor = PlayerColor.BLACK;
	protected String gameName;
	private boolean gameOver = false;
	protected HiveBasicGameBoard gameBoard;
	protected U ai = null;
	protected long time;
	protected int moves;

	public HiveAI(String username, String password, PlayerColor playerColor, String gameName ){
		this.username = username;
		this.password = password;
		//this tells us whether we need to initialize a game
		this.playerColor = playerColor;
		this.gameName = gameName;
	}
	
	protected abstract U createAI();
	
	protected abstract void performIntroduce( W move );

	protected abstract void performFirstMove( W move );
	
	protected abstract void performMove( W move );

	public void setServerHandler(ServerHandler handler){
		this.handler = handler;
	}

	public boolean isGameOver(){
		return (gameBoard.isGameOver() || this.gameOver);
	}

	@Override
	public void awaitLogin() {
		System.err.println("SwarmAI.awaitLogin");
		try {
			command = new ServerCommand( handler );
		} catch( IOException ex ) {
			System.err.println( "ERROR. Connection broken." );
			command = null;
			return;
		}
		command.login( username, password );
		System.err.println("AI now tries to log in");
	}


	@Override
	public void loginName(String name) {
		this.loginName = name;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void gameCreated( String name ) {
		System.err.println("AI received GAME_CREATED "+name+" message");
		if(name.equals(gameName) && (this.playerColor.equals(PlayerColor.BLACK))){//wenn dies das Spiel ist, dem wir beitreten sollen, dann machen wir das jetzt
			command.joinGame(name);
			System.err.println("AI now tries to join game "+name);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void success( Success success ) {
		switch( success ) {
		case CreateGame://the ai has successfully created a game
			break;
		case JoinGame:
			break;
		case Login://the ai has successfully joined, if we ever want to use the chat then this is the place to say something
			System.err.println("Successfully logged in");
			createAI();
			//starte ein Spiel falls wir weiss sind
			if(this.playerColor.equals(PlayerColor.WHITE)) {
				System.err.println("AI ist weiß und legt jetzt das Spiel " + this.gameName + " an.");
				command.createGame( this.gameName );
			} else {
				// We are a black player. Try to join the game submitted.
				command.joinGame( gameName );
			}
			break;
		default:
			throw new IllegalArgumentException( "Unsupported success message" );
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void failed( Failed failed ) {
		switch( failed ) {
		case Create:
			if( playerColor == PlayerColor.BLACK )
				throw new IllegalStateException( "Black player failed in creating a game. White should create game anyway." );
			System.err.println( "Could not create the game " + gameName + "." );
			command.logout();
			break;
		case Join:
			System.err.println( "Could not join the game " + gameName + " as " + playerColor + " player." );
			command.logout();
			break;
		case Introduce:
			//darf im echten Spiel nicht geschehen sonst verliert die KI
			System.err.println( "Failed introduce!" );
			command.logout();
			break;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void joinGame( String gameName, String playerName ) {
		if( gameName.equals( this.gameName ) && playerName.equals( this.loginName ) && playerColor == PlayerColor.BLACK ) {
			// we have successfully joioned a game as black player.
			System.out.println( "KI: black has successfully joined a game." );
		} else if(this.playerColor.equals(PlayerColor.WHITE) && this.gameName.equals(gameName) && !playerName.equals(this.loginName)){
			//a player has joined the game, if this is our game and we are the white player we need to start the game
			//start game
			command.startGame( gameName );
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void startGame( String game, String playerName ) {
		System.err.println( "START GAME received" );
		if( game.equals( this.gameName ) ) {
			// the game we are in has started.
			if( playerColor == PlayerColor.WHITE ) {
				// we have to perform our first move
				if( loginName.equals( playerName ) ) {
					//der weisse Spieler hat das Spiel erfolgreich gestartet, was uns der Server mit dieser Nachricht mitteilt
					//nun den ersten Zug berechnen und abschicken
					//W movement = ai.aiAction();
					W movement = ai.getWhiteFirst();
					performFirstMove( movement );
				} else {
					throw new IllegalStateException( "White KI player is not the first player in game!" );
				}
			} else {
				if( playerName.equals( loginName ) ) {
					throw new IllegalStateException( "Black KI player is first player in game!" );
				}
				System.out.println( "KI: Game started. Waiting for first white move." );
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void hiveWhiteFirst( String game, String player, PieceType pieceType ) {
		if(this.gameName.equals(game)) {
			//der weisse Spieler hat seinen ersten Stein gesetzt
			//nsai.introduceLocal( pieceType );
			//gameBoard.introduce( pieceType );//PlayerColor hat sich geaendert
			ai.introduce( pieceType ); // do this to gameboard is enough. however, delegate the ai may want to do something


			if( player.equals( loginName ) ) {
				// move from the other part
				if( playerColor == PlayerColor.BLACK )
					throw new IllegalStateException( "Black AI has performed white first move!" );
			} else {
				if( playerColor == PlayerColor.WHITE )
					throw new IllegalStateException( "White KI has not performed first move!" );
				
				System.out.println( "KI: computes the first black move." );
			
				long start = System.nanoTime();
				//W movement = ai.aiAction(); //.getBlackFirst();
				W movement = ai.getBlackFirst();
				long end = System.nanoTime();
				time += (end-start);
				moves++;
				performIntroduce( movement );
			}
		}
	}//hiveWhiteFirst

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void hiveBlackFirst( String game, String player, PieceType pieceType, HexDirection dir ) {
		if( game.equals( this.gameName ) ) {
			ai.introduce( pieceType, 0, dir );
			//zki.introduce( pieceType, 0, dir );
			if( player.equals( loginName ) ) {
				// we have introduced the first black piece
				if( playerColor == PlayerColor.WHITE )
					throw new IllegalStateException( "White KI has performed black first move!" );
				// have been black, so do nothing and wait for the next move of white player.
				
			} else {
				if( playerColor == PlayerColor.BLACK )
					throw new IllegalStateException( "Black has not performed first black move!" );
				// we are white, so compute next move
					System.out.println( "KI: computes new white move." );
					//aiMove();
					long start = System.nanoTime();
					W movement = ai.aiAction();
					long end = System.nanoTime();
					time += (end-start);
					moves++;
					performMove( movement );
					
			}
		}
	}//hiveBlackFirst

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void hiveIntroducePiece( String game, String player, PieceType pieceType, int anchorPiece, HexDirection hexDirection ) {
		if( game.equals( this.gameName ) ) {
			ai.introduce( pieceType, anchorPiece, hexDirection );	// delegate
			if( gameBoard.isGameOver() ) {
				System.out.println( "Game is over." );
				if( gameBoard.getWinner() == playerColor )
					System.out.println( "We have won!" );
				else
					System.out.println( "We haven't won." );
			} else if( !player.equals( loginName ) && gameBoard.getCurrentPlayer() == playerColor ) {
				// if the other player has done something, we are next
				long start = System.nanoTime();
				W movement = ai.aiAction();
				long end = System.nanoTime();
				time += (end - start);
				moves++;
				performMove( movement );
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void hiveMovePiece( String game, String player, int piece, int anchorPiece, HexDirection hexDirection ) {
		if( game.equals( this.gameName ) ) {
			ai.move( piece, anchorPiece, hexDirection ); // Delegate the event
			if( gameBoard.isGameOver() ) {
				
				System.out.println( "Game is over." );
				if( gameBoard.getWinner() == playerColor )
					System.out.println( "We have won!" );
				else
					System.out.println( "We haven't won." );
			} else if( !player.equals( loginName ) && gameBoard.getCurrentPlayer() == playerColor ) {
				long start = System.nanoTime();
				W action = ai.aiAction();
				long end = System.nanoTime();
				time += (end-start);
				moves++;
				performMove( action );
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void gameOver( String gameName, int result ) {
		if( gameName.equals( this.gameName ) ) {
			if( result == 0 )
				System.out.println("Spieler weiss hat das Spiel gewonnen.");
			else if( result == 1 )
				System.out.println("Spieler schwarz hat das Spiel gewonnen.");
			else
				System.out.println("Kein Spieler hat das Spiel gewonnen. Unentschieden!");
			this.gameOver = true;
			
			if( gameBoard.isGameOver() != true ) {
				System.out.println( "Irregular game over. Quit now." );
			}
			command.logout();
		}
	}

	@Override
	public void logout( String string ) {
		if( string == null ) {
			// this user has been logged out! reset everything to the start situation
			this.command = null;
			//this.nsai = null;
			handler.setConnectionClosed();
			handler = null;
			this.loginName = "";
			this.password = "";
			System.out.println( "Moves: " + moves );
			System.out.println( "Computation time: " + Formatter.formatTimeUnit( time, Formatter.TimeUnits.NanoSeconds ) );
			if( moves > 0 )
				System.out.println( "Time per move: " + Formatter.formatTimeUnit( time/moves, Formatter.TimeUnits.NanoSeconds ) );
		}
	}

	public void sendLogout(){
		command.logout();
	}

	@Override
	public void chatInfo(String message) {}

	@Override
	public void chatJoinChannel(String channel, String who) {}

	@Override
	public void chatLeaveChannel(String channel, String name) {}

	@Override
	public void chatMessage(String channel, String from, String message) {}
    
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void games( String[] parts ) {}

